import os
import signal
import sys
import time
from multiprocessing import Queue, Lock, JoinableQueue
from multiprocessing.managers import BaseManager
from typing import Any

import _pytest

from pytest_mproc.config import MPManagerConfig, RoleEnum

_getscopeitem_orig = _pytest.fixtures.FixtureRequest._getscopeitem


def _getscopeitem_redirect(self, scope):
    if scope == "global" or scope == "node":
        return self._pyfuncitem.getparent(_pytest.main.Session)
    else:
        return _getscopeitem_orig(self, scope)

_pytest.fixtures.scopes.insert(0, "node")
_pytest.fixtures.scopes.insert(0, "global")
_pytest.fixtures.scopenum_function = _pytest.fixtures.scopes.index("function")
_pytest.fixtures.scope2props['global'] = ()
_pytest.fixtures.scope2props['node'] = ()
_pytest.fixtures.FixtureRequest._getscopeitem = _getscopeitem_redirect


class FixtureManager(BaseManager):
    class Value:

        def __init__(self, value: Any):
            self._value = value

        def value(self):
            return self._value

    def __init__(self, host, port, is_master: bool):
        self._fixtures = {}
        self._is_master = is_master
        if not is_master:
            self.__class__.register("get_start_gate")
            self.__class__.register("get_fixtures")
            self.__class__.register("register_fixtures")
            self.__class__.register("wait_lock")
            self.__class__.register("is_test_q_exhausted")
        else:
            self._start_gate = Queue()
            self._gate = Queue()
            self._wait_lock = Lock()
            self._wait_lock.acquire()
            self.__class__.register("get_start_gate", lambda: self._start_gate)
            self.__class__.register("get_fixtures", self._get_fixtures)
            self.__class__.register("register_fixtures", self._register_fixtures)
            self.__class__.register("wait_lock", lambda: self._wait_lock)
            self.__class__.register("is_test_q_exhausted", lambda: self.Value(self._test_q_exhausted))
        super().__init__((host, port), authkey=b'pass')
        self._on_hold = True
        self._test_q_exhausted = False

    def _get_fixtures(self):
        v =self._gate.get()
        self._gate.put(v)
        return self._global_fixturedefs

    def _register_fixtures(self, fixturedefs):
        self._gate.put(True)
        self._global_fixturedefs = fixturedefs



class Node(_pytest.main.Session):

    class Manager(FixtureManager):
        fixturedefs = {}

        def __init__(self, config: MPManagerConfig, force_as_client: bool = False):
            # scope to node: only run on localhost with random port
            as_master = not force_as_client and (config.role in [RoleEnum.MASTER, RoleEnum.COORDINATOR])
            super().__init__(config.node_mgr_host, config.node_mgr_port, as_master)
            if as_master and not force_as_client:
                super().start()
            else:
                tries_remaining = 3
                while tries_remaining > 0:
                    try:
                        super().connect()
                        break
                    except (ConnectionRefusedError, ConnectionAbortedError, ConnectionError, ConnectionResetError):
                        tries_remaining -= 1
                        print(f"Connection error when connecting worker;  {tries_remaining} tries left")
                        time.sleep(1)


class Global(_pytest.main.Session):

    class Manager(FixtureManager):

        fixturedefs = {}

        def __init__(self, config: MPManagerConfig, force_as_client: bool = False):
            self._satellite_worker_count = 0
            self._test_q_exhausted = False
            self._mpconfig = config
            self._global_fixturedefs = {}
            if config.role != RoleEnum.MASTER or force_as_client:
                # client
                Global.Manager.register("register_satellite")
            else:
                # server:
                Global.Manager.register("register_satellite", self._register_satellite)
            super().__init__(config.global_mgr_host, config.global_mgr_port, config.role == RoleEnum.MASTER)
            if config.role == RoleEnum.MASTER and not force_as_client:
                super().start()
            else:
                def timeout(*args):
                    raise TimeoutError()

                signal.signal(signal.SIGALRM, timeout)
                tries_remaining = 60
                if config.role != RoleEnum.WORKER:
                    os.write(sys.stderr.fileno(), b"Waiting for server...\n")
                while tries_remaining:
                    try:
                        signal.alarm(1)
                        super().connect()
                        signal.alarm(0)
                        break
                    except (ConnectionResetError, ConnectionError, ConnectionAbortedError, ConnectionRefusedError) as e:
                        if config.role != RoleEnum.WORKER:
                            os.write(sys.stderr.fileno(), b".")
                        tries_remaining -= 1
                        if tries_remaining == 0:
                            os.write(sys.stderr.fileno(), b"\n")
                            raise Exception(f"Failed to connect to server {config.global_mgr_host} port " +
                                            f"{config.global_mgr_port}: {str(e)}")
                        time.sleep(0.5)
                    except TimeoutError:
                        tries_remaining -= 1
                        time.sleep(0.5)

                if config.role != RoleEnum.WORKER:
                    os.write(sys.stderr.fileno(), b"\nConnected\n")

        def _register_satellite(self, count: int):
            self._satellite_worker_count += count
            if self._test_q_exhausted:
                for _ in range(count):
                    self._test_q.put(None)

    def __init__(self, config):
        super().__init__(config)

    def setup(self):
        raise Exception("UNIMPL")

    def gethookproxy(self, fspath):
        raise Exception("UNIMPL")

    def isinitpath(self, path):
        return path in self.session._initialpaths

    def collect(self):
        raise Exception("UNIMPL")


_pytest.fixtures.scopename2class.update({"global": Global, "node": Node})



import os
import sys
from typing import AnyStr, Any


def is_degraded():
    return os.environ.get("PYTEST_MPROC_DISABLED")


class BasicReporter:

    def write(self, text: AnyStr, *args: Any, **kargs: Any):
        os.write(sys.stderr.fileno(), text.encode('utf-8'))



from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional

from pytest_mproc import find_free_port


class RoleEnum(Enum):
    MASTER = "MASTER"
    COORDINATOR = "COORDINATOR"
    WORKER = "WORKER"


@dataclass
class MPManagerConfig:
    role: RoleEnum
    num_processes: int
    global_mgr_host: str
    global_mgr_port: int
    node_mgr_host: str = '127.0.0.1'
    node_mgr_port: int = find_free_port()
    global_fixtures: Optional[Dict] = None
    node_fixtures: Optional[Dict] = None



"""
This file contains some standard pytest_* plugin hooks to implement multiprocessing runs
"""
import getpass
import inspect
import os
import shutil
import tempfile
from contextlib import contextmanager, suppress
from pathlib import Path
from traceback import format_exc

import _pytest

from pytest_mproc import find_free_port
from pytest_mproc.config import MPManagerConfig, RoleEnum
from pytest_mproc.fixtures import Node, Global

import pytest
import _pytest.terminal
import socket

from multiprocessing import cpu_count
from pytest_mproc.coordinator import Coordinator
from pytest_mproc.utils import is_degraded, BasicReporter


def _get_ip_addr():
    try:
        return socket.gethostbyname(socket.gethostname())
    except Exception:
        print(">>> Cannot get ip address of host.  Return 127.0.0.1 (localhost)")
        return "127.0.0.1"


def parse_numprocesses(s):
    """
    A little bit of processing to get number of parallel processes to use (since "auto" can be used to represent
    # of cores on machine)
    :param s: text to process
    :return: number of parallel worker processes to use
    """
    try:
        if s.startswith("auto"):
            if '*' in s:
                multiplication_factor = int(s.rsplit('*', 1)[-1])
            elif s == "auto":
                multiplication_factor = 1
            else:
                raise Exception("Error: --cores argument must be an integer value or auto or auto*<int factor>")
            return cpu_count() * multiplication_factor
        else:
            return int(s)
    except ValueError:
        raise Exception("Error: --cores argument must be an integer value or \"auto\" or \"auto*<int factor>\"")


@pytest.mark.tryfirst
def pytest_addoption(parser):
    """
    add options to given parser for this plugin
    """
    group = parser.getgroup("pytest_mproc", "better distributed testing through multiprocessing")
    group._addoption(
        "--cores",
        dest="mproc_numcores",
        metavar="mproc_numcores",
        action="store",
        type=parse_numprocesses,
        help="you can use 'auto' here to set to the number of  CPU cores on host system",
    )
    group._addoption(
            "--disable-mproc",
            dest="mproc_disabled",
            metavar="mproc_disabled",
            action="store",
            type=bool,
            help="disable any parallel mproc testing, overriding all other mproc arguments",
        )
    group._addoption(
        "--as-server",
        dest="mproc_server_port",
        metavar="mproc_server_port",
        action="store",
        type=int,
        help="port on which you wish to run server (for multi-host runs only)"
    )
    group._addoption(
        "--as-client",
        dest="mproc_client_connect",
        metavar="mproc_client_connect",
        action="store",
        type=str,
        help="host:port specification of master node to connect to as client"
    )


@pytest.mark.tryfirst
def pytest_cmdline_main(config):
    """
    Called before "true" main routine.  This is to set up config values well ahead of time
    for things like pytest-cov that needs to know we are running distributed

    Mostly taken from other implementations (such as xdist)
    """
    if config.option.collectonly:
        return

    reporter = BasicReporter()
    worker = getattr(config.option, "mproc_worker", None)
    mproc_server_port = getattr(config.option, 'mproc_server_port', None)
    mproc_client_connect = getattr(config.option, "mproc_client_connect", None)
    if mproc_client_connect and mproc_server_port:
        raise pytest.UsageError("Cannot specify both -as-master and --as-client at same time")
    config.option.numprocesses = config.option.mproc_numcores  # this is what pycov uses to determine we are distributed
    # We are a:
    #   WORKER no matter what if mproc_worker attr is set (per worker_main routine, i.e. started as a worker thread)
    #   MASTER if server port is specified on command line OR if neither server port nor client connection is specified
    #      on command line. In the latter case, the request is simply to run all tests on localhost, not remtoe workers
    #   COORDINATOR if client connect is specified
    role = RoleEnum.WORKER if worker is not None else \
        RoleEnum.MASTER if (mproc_server_port is not None or
                            (mproc_server_port is None and mproc_client_connect is None)) else \
        RoleEnum.COORDINATOR
    if role == RoleEnum.MASTER:
        reporter.write(f"Running as {role}: {_get_ip_addr()}:{mproc_server_port}\n", green=True)

    # validation
    if getattr(config.option, "mproc_numcores", None) is None or is_degraded() or getattr(config.option, "mproc_disabled"):
        config.option.mpconfig = MPManagerConfig(
            global_mgr_host='127.0.0.1',
            global_mgr_port=find_free_port(),
            role=role,
            num_processes=0)
        config.mproc_global_manager = Global.Manager(config.option.mpconfig)
        config.mproc_node_manager = Node.Manager(config.option.mpconfig)
        reporter.write(">>>>> no number of cores provided or running in environment unsupportive of parallelized testing, "
              "not running multiprocessing <<<<<\n", yellow=True)
        return
    elif hasattr(config.option, "mpconfig"):
        mpconfig = config.option.mpconfig
    elif mproc_server_port:
        if config.option.numprocesses < 0:
            raise pytest.UsageError("Number of cores must be greater than or equal to zero when running as a master")
        mpconfig = MPManagerConfig(global_mgr_host=_get_ip_addr(),
                                   global_mgr_port=mproc_server_port,
                                   role=role,
                                   num_processes=config.option.mproc_numcores)
    elif mproc_client_connect:
        if config.option.numprocesses < 1:
            raise pytest.UsageError("Number of cores must be 1 or more when running as client")
        try:
            host, port = mproc_client_connect.rsplit(':', maxsplit=1)
        except Exception:
            raise pytest.UsageError("--as-client must be specified in form '<host>:<port>' of the master node")
        try:
            port = int(port)
        except ValueError:
            raise pytest.UsageError("When specifying connection as client, port must be an integer value")

        mpconfig = MPManagerConfig(role=role,
                                   global_mgr_host=host,
                                   global_mgr_port=port,
                                   num_processes=config.option.mproc_numcores)
    else:
        if config.option.numprocesses < 1:
            raise pytest.UsageError("Number of cores must be 1 or more when running on a single node")
        mpconfig = MPManagerConfig(role=role,
                                   global_mgr_host='127.0.0.1',
                                   global_mgr_port=find_free_port(),
                                   num_processes=config.option.mproc_numcores)
    config.option.mpconfig = mpconfig
    # tell xdist not to run, (and BTW setting numprocesses is enough to tell pycov we are distributed)
    config.option.dist = "no"
    val = config.getvalue
    if not val("collectonly"):
        usepdb = config.getoption("usepdb")  # a core option
        if val("dist") != "no" and usepdb:
            raise pytest.UsageError(
                "--pdb is incompatible with distributing tests."
            )  # noqa: E501
    config.coordinator = None
    if role in [RoleEnum.MASTER, RoleEnum.COORDINATOR]:
        # in main thread,
        # instantiate coordinator here and start to kick off processing on workers early, so they can
        # process config info in parallel to this thread
        config.coordinator = Coordinator(mpconfig)
    config.mproc_global_manager = Global.Manager(mpconfig)
    config.mproc_node_manager = Node.Manager(mpconfig)
    if role != RoleEnum.WORKER:
        config.coordinator.start(mpconfig)
    elif hasattr(config.option, "connect_sem"):
        config.option.connect_sem.release()


def pytest_sessionstart(session):
    with suppress(Exception):
        reporter = session.config.pluginmanager.getplugin('terminalreporter')
        if reporter:
                reporter._session = session


@pytest.mark.trylast
def pytest_configure(config):
    if getattr(config.option, "mproc_numcores", None) is None or is_degraded() or getattr(config.option, "mproc_disabled"):
        return  # return of None indicates other hook impls will be executed to do the task at hand
    # tell xdist not to run, (and BTW setting numprocesses is enough to tell pycov we are distributed)
    config.option.dist = "no"


@pytest.mark.tryfirst
def pytest_fixture_setup(fixturedef, request):
    my_cache_key = fixturedef.cache_key(request)
    if request.config.option.mpconfig.role == RoleEnum.WORKER:
        if request.config.option.mpconfig.global_fixtures is None:
            client = request.config.mproc_global_manager # Global.Manager(request.config.option.mpconfig, force_as_client=True)
            request.config.option.mpconfig.global_fixtures = client.get_fixtures()
        if request.config.option.mpconfig.node_fixtures is None:
            client = request.config.mproc_node_manager # Node.Manager(request.config.option.mpconfig, force_as_client=True)
            request.config.option.mpconfig.node_fixtures = client.get_fixtures()
        if fixturedef.scope == 'global':
            fixturedef.cached_result = (request.config.option.mpconfig.global_fixtures.get(fixturedef.argname),
                                        my_cache_key,
                                        None)
            return request.config.option.mpconfig.global_fixtures.get(fixturedef.argname)
        elif fixturedef.scope == 'node':
            fixturedef.cached_result = (request.config.option.mpconfig.node_fixtures.get(fixturedef.argname),
                                        my_cache_key,
                                        None)
            return request.config.option.mpconfig.node_fixtures.get(fixturedef.argname)
        else:
            return

    if not getattr(request.config.option, "mproc_numcores") or fixturedef.scope not in ['node', 'global']:
        return
    if not hasattr(request.config, "generated_fixtures"):
        request.config.generated_fixtures = []
    if request.config.option.mpconfig.global_fixtures is None:
        request.config.option.mpconfig.global_fixtures = {}
    if request.config.option.mpconfig.node_fixtures is None:
        request.config.option.mpconfig.node_fixtures = {}
    if fixturedef.scope == 'node':
        assert request.config.option.mpconfig.node_fixtures.get(fixturedef.argname) is not None
        fixturedef.cached_result = (request.config.option.mpconfig.node_fixtures.get(fixturedef.argname),
                                    my_cache_key,
                                    None)
        return fixturedef.cached_result[0]
    elif fixturedef.scope == 'global':
        assert request.config.option.mpconfig.global_fixtures.get(fixturedef.argname) is not None
        fixturedef.cached_result = (request.config.option.mpconfig.global_fixtures.get(fixturedef.argname),
                                    my_cache_key,
                                    None)
        return fixturedef.cached_result[0]


def process_fixturedef(fixturedef, request):
    generated = request.config.generated_fixtures
    global_fixtures = request.config.option.mpconfig.global_fixtures
    node_fixtures = request.config.option.mpconfig.node_fixtures
    config = request.config
    mpconfig: MPManagerConfig = config.option.mpconfig
    if not fixturedef or fixturedef.scope not in ['global', 'node']:
        return
    name = fixturedef.argname
    assert mpconfig.role != RoleEnum.WORKER
    if mpconfig.role == RoleEnum.MASTER and fixturedef.scope == 'global' and name not in global_fixtures:
        global_fixtures[name] = _pytest.fixtures.resolve_fixture_function(fixturedef, request)
        val = global_fixtures[name](*fixturedef.argnames)
        if inspect.isgenerator(val):
            generated.append(val)
            val = next(val)
        mpconfig.global_fixtures[name] = val
    if mpconfig.role in [RoleEnum.MASTER, RoleEnum.COORDINATOR] and fixturedef.scope == 'node'\
            and name not in node_fixtures:
        node_fixtures[name] = _pytest.fixtures.resolve_fixture_function(fixturedef, request)
        val = node_fixtures[name](*fixturedef.argnames)
        if inspect.isgenerator(val):
            generated.append(val)
            val = next(val)
        mpconfig.node_fixtures[name] = val


def pytest_runtestloop(session):
    if session.config.option.collectonly:
        return
    reporter: _pytest.terminal.TerminalReporter = session.config.pluginmanager.getplugin('terminalreporter')
    if reporter:
        reporter.tests_count = len(session.items)
    if reporter is None:
        reporter = BasicReporter()
    if session.testsfailed and not session.config.option.continue_on_collection_errors:
        raise session.Interrupted("%d errors during collection" % session.testsfailed)

    if session.config.option.collectonly:
        if session.config.option.mpconfig.role in [RoleEnum.MASTER, RoleEnum.COORDINATOR]:
            if hasattr(session.config, "coordinator"):
                session.config.coordinator.kill()
        return  # should never really get here, but for consistency
    if len(session.items) == 0:
        return
    if getattr(session.config.option, "mproc_numcores", None) is None or is_degraded() or getattr(session.config.option, "mproc_disabled"):
        # return of None indicates other hook impls will be executed to do the task at hand
        # aka, let the basic hook handle it from here, no distributed processing to be done
        return
    mpconfig = session.config.option.mpconfig
    session.config.generated_fixtures = []
    if mpconfig.role != RoleEnum.WORKER:
        if mpconfig.global_fixtures is None:
            mpconfig.global_fixtures = {}
        if mpconfig.node_fixtures is None:
            mpconfig.node_fixtures = {}
        for item in session.items:
            if session.shouldfail:
                break
            if hasattr(item, "_request"):
                for name in item._fixtureinfo.argnames:
                    try:
                        if name not in mpconfig.global_fixtures and name not in mpconfig.node_fixtures:
                            fixturedef = item._fixtureinfo.name2fixturedefs.get(name, [None])[0]
                            if not fixturedef:
                                continue
                            for argname in fixturedef.argnames:
                                fixdef = item._fixtureinfo.name2fixturedefs.get(argname, [None])[0]
                                process_fixturedef(fixdef, item._request)

                            process_fixturedef(fixturedef, item._request)
                    except Exception as e:
                        session.shouldfail = f"Exception in fixture: {e.__class__.__name__} raised with msg '{str(e)}'" + \
                            f"{format_exc()}"
                        reporter.write(f">>> Fixture ERROR: {format_exc()}\n", red=True)
                        break

    if session.shouldfail:
        if mpconfig.role in [RoleEnum.MASTER, RoleEnum.COORDINATOR]:
            if hasattr(session.config, "coordinator"):
                session.config.coordinator.kill()
        raise session.Failed(session.shouldfail)
    elif mpconfig.role != RoleEnum.WORKER:
        try:
            if mpconfig.role == RoleEnum.MASTER:
                session.config.mproc_global_manager.register_fixtures(mpconfig.global_fixtures)
            session.config.mproc_node_manager.register_fixtures(mpconfig.node_fixtures)
        except Exception as e:
            import pickle
            for i, v in mpconfig.node_fixtures.items():
                reporter.write(f">>> {i}\n", red=True)
                reporter.write(f">>> {pickle.dumps(v)}\n", red=True)
            session.shouldfail = f"Exception in fixture: {e.__class__.__name__} raised with msg '{str(e)}'" + \
                f"{format_exc()}\n {mpconfig.node_fixtures}"
            reporter.write(f">>>> Error registering node/global level fixture(s): {str(e)}\n", red=True)
            if hasattr(session.config, "coordinator"):
                session.config.coordinator.kill()
            raise session.Failed(session.shouldfail)

    if mpconfig.role != RoleEnum.WORKER:
        if not session.config.getvalue("collectonly"):
            # main coordinator loop:
            with session.config.coordinator as coordinator:
                try:
                    coordinator.set_items(session.items)
                    coordinator.run(session)
                except Exception as e:
                    reporter.write(format_exc() + "\n", red=True)
                    reporter.write(f"\n>>> ERROR in run loop;  unexpected Exception\n {str(e)}\n\n", red=True)
                    return False
    elif mpconfig.role == RoleEnum.WORKER:
        worker = getattr(session.config.option, "mproc_worker")
        worker.test_loop(session)

    return True


def pytest_terminal_summary(terminalreporter: _pytest.terminal.TerminalReporter, exitstatus, config):
    if config.option.collectonly:
        return
    if config.option.mpconfig.role != RoleEnum.MASTER:
        terminalreporter.line(">>>>>>>>>> This is a satellite processing node. "
                              "Please see master node output for actual test summary <<<<<<<<<<<<",
                              yellow=True)


def pytest_sessionfinish(session):
    reporter = BasicReporter()
    if session.config.option.collectonly:
        return
    mpconfig = session.config.option.mpconfig
    if mpconfig.role in [RoleEnum.MASTER, RoleEnum.COORDINATOR]:
        if hasattr(session.config, "coordinator"):
            session.config.coordinator.kill()
    if mpconfig.role == RoleEnum.MASTER:
        try:
            session.config.mproc_global_manager.shutdown()
        except Exception as e:
            reporter.write(">>> INTERNAL Error shutting down mproc manager\n", red=True)
    if mpconfig.role in [RoleEnum.MASTER, RoleEnum.COORDINATOR]:
        try:
            session.config.mproc_node_manager.shutdown()
        except Exception as e:
            reporter.write(">>> INTERNAL Error shutting down mproc manager\n", red=True)
    generated = getattr(session.config, "generated_fixtures", [])
    errors = []
    for item in generated:
        try:
            try:
                next(item)
                raise Exception(f"Fixture generator did not stop {item}")
            except StopIteration:
                pass
        except Exception as e:
            reporter.write(f">>> Error in exit of fixture context {e}\n", red=True)
            errors.append(e)

    if errors:
        raise pytest.UsageError(">>> One or more global or node-level fixtures threw an Exception on exit from context") \
            from errors[0]

################
# Process-safe temp dir
###############


class TmpDirFactory:

    """
    tmpdir is not process/thread safe when used in a multiprocessing environment.  Failures on setup can
    occur (even if infrequently) under certain rae conditoins.  This provides a safe mechanism for
    creating temporary directories utilizng s a global-scope fixture
    """

    def __init__(self):
        self._root_tmp_dir = tempfile.mkdtemp(prefix=f"pytest_mproc-{getpass.getuser()}-{os.getpid()}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with suppress(Exception):
            shutil.rmtree(self._root_tmp_dir)

    @contextmanager
    def create_tmp_dir(self, cleanup_immediately: bool = True):
        """
        :param cleanup_immediately: if True, rm the directory and all contents when associated fixture is no longer
           in use, otherwise wait until end of test session when everything is cleaned up
        :return: newly create temp directory unique across all Process's
        """
        tmpdir = tempfile.mkdtemp(dir=self._root_tmp_dir)
        try:
            yield Path(tmpdir)
        finally:
            if cleanup_immediately:
                with suppress(Exception):
                    shutil.rmtree(tmpdir)


@pytest.fixture(scope='node')
def mp_tmpdir_factory():
    """
    :return: a factory for creating unique tmp directories, unique across all Process's
    """
    with TmpDirFactory() as factory:
        yield factory


@pytest.fixture()
def mp_tmp_dir(mp_tmpdir_factory: TmpDirFactory):
    # tmpdir is not thread safe and can fail on test setup when running on a highly loaded very parallelized system
    # so use this instead
    with mp_tmpdir_factory.create_tmp_dir("PYTEST_MPROC_LAZY_CLEANUP" not in os.environ) as tmp_dir:
        yield tmp_dir



"""
This package contains code to coordinate execution from a main thread to worker threads (processes)
"""
import os
import resource
import signal
import sys
import time
from typing import List, Any

from _pytest.config import _prepareconfig
from multiprocessing import Process, Queue, Semaphore, JoinableQueue

from pytest_mproc import resource_utilization
from pytest_mproc.config import MPManagerConfig, RoleEnum
from pytest_mproc.utils import BasicReporter
from pytest_mproc.worker import main as worker_main
from pytest_mproc.fixtures import Global


class Coordinator:
    """
    Context manager for kicking of worker Processes to conduct test execution via pytest hooks
    """

    MAX_SIMULTANEOUS_CONNECT = 25

    def __init__(self, config: MPManagerConfig):
        """
        :param num_processes: number of parallel executions to be conducted
        """
        self._num_processes = config.num_processes
        self._tests = []  # set later
        self._role = config.role
        # test_q is for sending test nodeid's to worked
        # result_q is for receiving results as messages, exceptions, test status or any exceptions thrown
        self._count = 0
        self._rusage = []
        self._session_start_time = time.time()
        self._test_q: JoinableQueue = None  # set on call to start()
        self._result_q: JoinableQueue = None  # set on call to start()
        self._end_sem = Semaphore(0)
        self._reporter = BasicReporter()

    @staticmethod
    def _write_sep(s, txt):
        """
        write out text to stdout surrounded by repeated character

        :param s: character to repeat on either side of given text
        :param txt: text to by surrounded
        """
        sep_total = max((70 - 2 - len(txt)), 2)
        sep_len = sep_total // 2
        sep_extra = sep_total % 2
        out = '%s %s %s\n' % (s * sep_len, txt, s * (sep_len + sep_extra))
        sys.stdout.write(out)

    def _output_summary(self, time_span: float, ucpu: float, scpu: float, unshared_mem: float):
        """
        Output the summary of test execution
        """
        self._write_sep('=', "STATS")
        sys.stdout.write("User CPU, System CPU utilization, Add'l memory during run\n")
        sys.stdout.write("---------------------------------------------------------\n")
        for index, (count, duration, ucpu, scpu, mem) in enumerate(self._rusage):
            sys.stdout.write(f"Process Worker-{index} executed {count} tests in {duration:.2f} " +\
                             f"seconds; User CPU: {ucpu:.2f}%, Sys CPU: {scpu:.2f}%, " +\
                             f"Mem consumed: {unshared_mem/1000.0}M\n")
        sys.stdout.write("\n")
        sys.stdout.write(
            f"Process Coordinator executed in {time_span:.2f} seconds. " +
            f"User CPU: {ucpu:.2f}%, Sys CPU: {scpu:.2f}%, " +
            f"Mem consumed: {unshared_mem/1000.0}M\n"
        )
        length = sum([1 if not isinstance(t, list) else len(t) for t in self._tests])
        if self._count != length and self._role == RoleEnum.MASTER:
            self._write_sep('!', "{} tests unaccounted for {} out of {}".format(length - self._count,
                            self._count, length))
        sys.stdout.flush()

    def _process_worker_message(self, hook, typ, data, ):
        """
        Process a message (as a worker) from the coordinating process

        :param typ: the kind of messsage
        :param data: payload for the message to be processed
        """
        try:
            # TODO: possibly have more structured communications, but for now
            # there are limited types of messages, so sendo message, <data> tuple
            # seems to be fine?
            if typ == 'test_status':
                report = data
                if report.when == 'call' or (report.when == 'setup' and not report.passed):
                    self._count += 1
                hook.pytest_runtest_logreport(report=report)
                if report.failed:
                    sys.stdout.write("\n%s FAILED\n" % report.nodeid)
            elif typ == 'exit':
                # process is complete, so close it and set to None
                try:
                    index, worker_count, exitstatus, duration, rusage = data  # which process and count of tests run
                    time_span, ucpu, scpu, unshared_mem = rusage
                    self._rusage.append((worker_count, time_span, ucpu, scpu, unshared_mem))
                except:
                    pass
                    # self._reporter.write(f"Unable to process rusage: {data}\n")
            elif typ == 'error_message':
                error_msg_text = data
                sys.stdout.write("{}\n".format(error_msg_text))
            elif typ == 'exception':
                # reraise any exception from workers
                raise Exception("Exception in worker process: %s" % str(data)) from data

        except Exception as e:
            import traceback
            traceback.print_exc()
            sys.stdout.write("INTERNAL_ERROR> %s\n" % str(e))

    def read_results(self, hook):
        items = self._result_q.get()
        while items is not None:
            if isinstance(items, Exception):
                raise Exception
            for kind, data in items:
                self._process_worker_message(hook, kind, data)
            items = self._result_q.get()

    def do_work(self, mpconfig: MPManagerConfig, start_sem: Semaphore, end_sem: Semaphore):
        processes : List[Process] = []
        try:
            connect_sem: List[Semaphore] = [Semaphore(0) for i in range(mpconfig.num_processes)]
            batch_index = 0
            for index in range(mpconfig.num_processes):
                args = mpconfig.__dict__
                args.update({'role': RoleEnum.WORKER})
                worker_config = MPManagerConfig(**args)
                proc = Process(target=worker_main, args=(index, worker_config, mpconfig.role != RoleEnum.MASTER,
                                                         connect_sem[index], self._reporter, start_sem, self._test_q,
                                                         self._result_q))
                proc.start()
                if index % self.MAX_SIMULTANEOUS_CONNECT == self.MAX_SIMULTANEOUS_CONNECT-1 or index == mpconfig.num_processes-1:
                    for j in range(batch_index, index+1):
                        connect_sem[j].acquire()
                        connect_sem[j].release()
                        self._reporter.write(f"Worker-{j}: Connection established\n")
                    batch_index = index + 1
                processes.append(proc)

            def kill_all(*args, **kargs):
                self._reporter.write("Termination signal received;  killing all workers...")
                for _ in range(self._num_processes):
                    self._test_q.put(None)
                    self._test_q.join()
                for proc in processes:
                    proc.terminate()
                    proc.join(timeout=2)
                    processes.remove(proc)

            signal.signal(signal.SIGTERM, kill_all)
            start_sem.release()
            end_sem.acquire()
            for index, proc in enumerate(processes):
                proc.join(timeout=3)

            self._result_q.put(None)
            # self._result_q.join()
            self._result_q.close()
        except Exception as e:
            import traceback
            self._reporter.write(f">>>>>>>>>>> ERROR in worker manager process: str{e} \n{traceback.format_exc()}\n")
            for index, proc in enumerate(processes):
                if proc:
                    os.kill(processes[index].pid, signal.SIGKILL)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._worker_manager_process.join(timeout=2)
        except Exception:
            self.kill()

    def populate_test_queue(self, tests: List[Any], end_sem):
        try:
            count = 0
            for test in tests:
                # Function objects in pytest are not pickle-able, so have to send string nodeid and
                # do lookup on worker side
                item = [t.nodeid for t in test] if isinstance(test, list) else [test.nodeid]
                self._test_q.put(item)
                count += 1
                if count % 20 == 0:
                    self._test_q.join()
            self._test_q.join()
            for _ in range(self._num_processes):
                self._test_q.put(None)
                self._test_q.join()
            self._test_q.close()
        finally:
            end_sem.release()

    def start(self, mpconfig: MPManagerConfig):
        """
        Start all worker processes

        :return: this object
        """
        client = Global.Manager(mpconfig, force_as_client=True)
        client.register_satellite(mpconfig.num_processes)
        self._test_q: JoinableQueue = JoinableQueue()
        self._result_q: JoinableQueue = JoinableQueue()
        start_sem = Semaphore(0)
        self._worker_manager_process = Process(target=self.do_work, args=(mpconfig, start_sem, self._end_sem))
        self._worker_manager_process.start()
        start_sem.acquire()
        start_sem.release()
        return self

    def set_items(self, tests):
        """
        :param tests: the items containing the pytest hooks to the tests to be run
        """
        grouped = [t for t in tests if getattr(t._pyfuncitem.obj, "_pytest_group", None)]
        self._tests = [t for t in tests if t not in grouped]
        groups = {}
        for g in grouped:
            name, priority = g._pyfuncitem.obj._pytest_group
            groups.setdefault((name, priority), []).append(g)
        for key in sorted(groups.keys(), key=lambda x: x[1]):
            self._tests.insert(0, groups[key])

    def kill(self):
        self._worker_manager_process.join()
        self._worker_manager_process.terminate()

    def run(self, session):
        """
        Populate test queue and continue to process messages from worker Processes until they complete

        :param session: Pytest test session, to get session or config information
        """
        start_rusage = resource.getrusage(resource.RUSAGE_SELF)
        start_time = time.time()
        mpconfig = session.config.option.mpconfig
        # we are the root node, so populate the tests
        if mpconfig.role == RoleEnum.MASTER:
            populate_tests_process = Process(target=self.populate_test_queue, args=(self._tests, self._end_sem))
            populate_tests_process.start()
        self.read_results(session.config.hook)  # only master will read results and post reports through pytest
        populate_tests_process.join(timeout=1)  # should never time out since workers are done
        end_rusage = resource.getrusage(resource.RUSAGE_SELF)
        time_span = time.time() - start_time
        time_span, ucpu, scpu, addl_mem_usg = resource_utilization(time_span=time_span,
                                                                   start_rusage=start_rusage,
                                                                   end_rusage=end_rusage)

        sys.stdout.write("\r\n")
        self._output_summary(time_span, ucpu, scpu, addl_mem_usg)
        self._worker_manager_process.join()


if __name__ == "__main__":
    args = sys.argv[1:]
    plugins_to_load = []
    config = _prepareconfig(args, plugins_to_load)
    config.hook.pytest_cmdline_main(config=config)



import os
import socket
import sys
import time
from multiprocessing import Queue, Semaphore, JoinableQueue

import pytest
from _pytest.config import _prepareconfig, ConftestImportFailure
import pytest_cov.embed
import resource

from pytest_mproc import resource_utilization
from pytest_mproc.config import MPManagerConfig, RoleEnum
from pytest_mproc.fixtures import Global
from pytest_mproc.utils import BasicReporter

if sys.version_info[0] < 3:
    # noinspection PyUnresolvedReferences
    from Queue import Empty
else:
    # noinspection PyUnresolvedReferences
    from queue import Empty

try:
    my_cov = None  # overriden later, after fork
    try:
        original_cleanup = pytest_cov.embed.cleanupgit
    except:
        original_cleanup = None

    def _cleanup(cov=None):
        # pytest_cov will do special things when tests use multiprocessing, however,
        # here we don't want that processing to take place.  (It will otherwise
        # spit out many lines of messaging all of which has no bearing on success of code coverage
        # collecion)
        if cov != my_cov and original_cleanup:
            original_cleanup(cov)


    pytest_cov.embed.cleanup = _cleanup
    pycov_present = True

except Exception:
    pycov_present = False


"""maximum time between reporting status back to coordinator"""
MAX_REPORTING_INTERVAL = 1.0  # seconds


class WorkerSession:
    """
    Handles reporting of test status and the like
    """

    def __init__(self, index, result_q: JoinableQueue, test_q: JoinableQueue,
                 test_generator, is_remote: bool, mpconfig, start_sem: Semaphore):
        self._mpconfig = mpconfig
        self._is_remote = is_remote
        self._result_q = result_q
        self._test_q = test_q
        self._index = index
        self._test_generator = test_generator
        self._start_sem = start_sem

        self._name = "worker-%d" % (index + 1)
        self._reports = []
        self._count = 0
        self._session_start_time = time.time()
        self._buffered_results = []
        self._buffer_size = 20
        self._timestamp = time.time()
        self._last_execution_time = time.time()
        self._resource_utilization = -1, -1, -1, -1

    def _put(self, kind, data, timeout=None):
        """
        Append test result data to queue, flushing buffered results to queue at watermark level for efficiency

        :param kind: kind of data
        :param data: test result data to publih to queue
        """
        if self._is_remote and kind == 'test_status':
            os.write(sys.stderr.fileno(), b'.')
        self._buffered_results.append((kind, data))
        if len(self._buffered_results) >= self._buffer_size or (time.time() - self._timestamp) > MAX_REPORTING_INTERVAL:
            self._flush(timeout)

    def _flush(self, timeout=None):
        """
        fluh buffered results out to the queue.
        """
        if self._buffered_results:
            if timeout is not None:
                self._result_q.put(self._buffered_results, timeout)
            else:
                self._result_q.put(self._buffered_results)
            # self._result_q.join()
            self._buffered_results = []
            self._timestamp = time.time()

    def test_loop(self, session):
        """
        This is where the action takes place.  We override the usual implementation since
        that doesn't support a dynamic generation of tests (whose source is the test Queue
        that it draws from to pick out the next test

        :param session:  Where the tests generator is kept
        """
        self._start_sem.acquire()
        self._start_sem.release()
        start_time = time.time()
        rusage = resource.getrusage(resource.RUSAGE_SELF)
        try:
            if pycov_present:
                global my_cov
                try:
                    my_cov = pytest_cov.embed.active_cov
                except:
                    my_cov = None

            if session.testsfailed and not session.config.option.continue_on_collection_errors:
                raise session.Failed("%d errors during collection" % session.testsfailed)

            if session.config.option.collectonly:
                return  # should never really get here, but for consistency
            for items in session.items_generator:
                # test item comes through as a unique string nodeid of the test
                # We use the pytest-collected mapping we squirrelled away to look up the
                # actual _pytest.python.Function test callable
                # NOTE: session.items is an iterator, as wet upon construction of WorkerSession
                for item_name in items:
                    item = session._named_items[item_name]
                    item.config.hook.pytest_runtest_protocol(item=item, nextitem=None)
                    # very much like that in _pytest.main:
                    try:
                        if session.shouldfail:
                            raise session.Failed(session.shouldfail)
                    except AttributeError:
                        pass  # some version of pytest do not have this attribute
                    if session.shouldstop:
                        raise session.Interrupted(session.shouldstop)
                    # count tests that have been run
                    self._count += 1
                    self._last_execution_time = time.time()
            self._flush()
        finally:
            end_usage = resource.getrusage(resource.RUSAGE_SELF)
            time_span = self._last_execution_time - start_time
            self._resource_utilization = resource_utilization(time_span=time_span,
                                                              start_rusage=rusage,
                                                              end_rusage=end_usage)

    @pytest.mark.tryfirst
    def pytest_collection_finish(self, session):
        """
        Invoked once pytest has collected all information about which tests to run.
        Those items are squirrelled away in a different attributes, and a generator
        is put in its place to draw from.

        :param session: the pytest session

        :return: the generator of tests to run
        """
        session._named_items = {item.nodeid: item for item in session.items}
        session.items_generator = self._test_generator
        return session.items

    def pytest_internalerror(self, excrepr):
        self._put('error_message', excrepr)

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_logreport(self, report):
        """
        report only status of calls (not setup or teardown), unless there was an error in those
        stash report for later end-game parsing of failures

        :param report: report to draw info from
        """
        self._put('test_status', report)
        return True

    @pytest.hookimpl(tryfirst=True)
    def pytest_sessionfinish(self, exitstatus):
        """
        output failure information and final exit status back to coordinator

        :param exitstatus: exit status of the test suite run
        """
        # timeouts on queues sometimes can still have process hang if there is a bottleneck, so use alarm
        self._put('exit', (self._index, self._count, exitstatus, self._last_execution_time - self._session_start_time,
                           self._resource_utilization))
        self._flush()
        return True


def main(index, mpconfig: MPManagerConfig, is_remote: bool, connect_sem: Semaphore, reporter, start_sem: Semaphore,
         test_q: JoinableQueue, result_q: JoinableQueue):
    """
    main worker function, launched as a multiprocessing Process

    :param index: index assigned to the worker Process
    """
    def generator(test_q: JoinableQueue):
        try:
            test = test_q.get()
            while test:
                test_q.task_done()
                yield test
                test = test_q.get()
        finally:
            test_q.task_done()

    args = sys.argv[1:]
    worker = WorkerSession(index, result_q, test_q, generator(test_q), is_remote, mpconfig, start_sem)
    try:

        try:
            # use pytest's usual method to prepare configuration
            # we have to re-generate the config in this worker Process, as we cannot pickle the more complex
            # _pytest.python.Function objects across the Queue interface.  This is a slight bit
            # of awkwardness and some overhead per process, but this overhead pales in significance
            # to xdist's use of rsync
            config = _prepareconfig(args, plugins=[])
        except ConftestImportFailure as e:
            worker._put(('exception', e))
            worker._flush()
        # unregister terminal (don't want to output to stdout from worker)
        # as well as xdist (don't want to invoke any plugin hooks from another distribute testing plugin if present)
        config.pluginmanager.unregister(name="terminal")
        # register our listener, and configure to let pycov-test knoew that we are a slave (aka worker) thread
        config.pluginmanager.register(worker, "mproc_worker")
        config.option.mproc_worker = worker
        # setup slave info;  this is important to have something defined for these fields
        # so that pycov can pickup on the distributed nature of the test run,
        # and that this a worker(slave) process
        workerinput = {'slaveid': "worker-%d" % index,
                       'workerid': "worker-%d" % index,
                       "workercount": mpconfig.num_processes,
                       "slavecount": mpconfig.num_processes,
                       'cov_master_host': socket.gethostname(),
                       'cov_slave_output': os.path.join(os.getcwd(), "worker-%d" % index),
                       'cov_master_topdir': os.getcwd()
                       }
        config.slaveinput = workerinput
        config.slaveoutput = workerinput
        mpconfig.role = RoleEnum.WORKER
        mpconfig.index = index
        config.option.mpconfig = mpconfig
        config.option.mpconfig.node_fixtures = None
        config.option.mpconfig.global_fixtures = None
        config.option.connect_sem = connect_sem
        assert(mpconfig.role == RoleEnum.WORKER)
        try:
            # and away we go....
            config.hook.pytest_cmdline_main(config=config)
        finally:
            config._ensure_unconfigure()
    except Exception as e:
        import traceback
        reporter.write(f"Exception in worker thread: {str(e)}\n")
        reporter.write(f"Exception in worker thread: {traceback.format_exc()}\n")
        worker._put('exception', e, timeout=3)
        worker._put('exit', (index, -1, -1, -1, (-1. -1, -1, -1)))
        worker._flush()
    finally:
        reporter.write(f"\nWorker-{index} finished\n")



import inspect
import socket
import sys
from contextlib import closing


def group(name, priority: int = 0):
    """
    Decorator for grouping tests or a class of tests
    :param name: unique name for group of tests to be serialized under execution
    """
    def decorator_group(object):
        if inspect.isclass(object):
            for method in [getattr(object, m) for m in dir(object) if inspect.isfunction(getattr(object, m)) and m.startswith('test')]:
                method._pytest_group = (name, priority)
        elif inspect.isfunction(object):
            object._pytest_group = (name, priority)
        else:
            raise Exception("group decorator can only decorate class or function object")
        return object
    return decorator_group


def resource_utilization(time_span: float, start_rusage, end_rusage):
    if time_span <= 0.001:
        return -1, -1, -1, end_rusage.ru_maxrss
    if sys.platform.lower() == 'darwin':
        # OS X is in bytes
        delta_mem = (end_rusage.ru_maxrss - start_rusage.ru_maxrss) / 1000.0
    else:
        delta_mem = (end_rusage.ru_maxrss - start_rusage.ru_maxrss)
    ucpu_secs = end_rusage.ru_utime - start_rusage.ru_utime
    scpu_secs = end_rusage.ru_stime - start_rusage.ru_stime
    return time_span, \
           (ucpu_secs / time_span) * 100.0, \
           (scpu_secs / time_span) * 100.0, \
           delta_mem


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


import os
from pathlib import Path

import pytest_mproc
import sys
# to pick up dummy_src (ensure in path):
sys.path.append(os.path.basename(__file__))
sys.path.append(os.path.join(os.path.basename(__file__), "..", "testcode"))

import pytest
from testcode.testsomething import Something

os.environ["PYTEST_MPROC_LAZY_CLEANUP"] = "1"  # only cleanup tempdirs at end of run


@pytest.fixture()
def some_fixture():
    return 1


@pytest_mproc.group("group", 1)
def test_group_m2():
    Something().group_m1()
    assert Something.proc_id == os.getpid()


@pytest_mproc.group("class_group")
class TestGrouped:

    proc_id = None

    def test1(self):
        TestGrouped.proc_id = os.getpid()

    def test2(self):
        assert TestGrouped.proc_id == os.getpid()


def test_some_alg1(global_fix, mp_tmpdir_factory, queue_fixture):
    assert global_fix == 42
    Something().some_alg1()


@pytest.mark.parametrize('data', ['a%s' % i for i in range(1000)])
def test_some_alg2(data, some_fixture, global_fix, mp_tmp_dir: Path):
    unique_path = str(mp_tmp_dir.joinpath("unique_file.txt"))
    assert not os.path.exists(unique_path),  "mp_tmp_dir did not produce a unique path as expected"
    with open(unique_path, 'w'):
        pass
    assert global_fix == 42
    Something().some_alg2(data)


def test_some_alg3():
    Something().some_alg3()

